from jupyter_server_synchronizer.extension import SynchronizerExtension


def _jupyter_server_extension_points():  # pragma: no cover
    return [
        {
            "module": "jupyter_server_synchronizer.extension",
            "app": SynchronizerExtension,
        }
    ]



from dataclasses import dataclass, fields
from typing import List, Union

from jupyter_client.manager import KernelManager


class KernelRecordConflict(Exception):
    """An exception raised when trying to merge two
    kernels that have conflicting data.
    """

    pass


@dataclass
class KernelRecord:
    """A dataclass that keeps a record of kernels maintained
    by Jupyter Server's synchronizer.

    Two records are considered equivalent if their unique identifiers
     are equal. In this case, calling
    `.update(...)` will merge the data of two records
    """

    # A Kernel record should have at least one field
    # that ends with `_id` in its keyname.
    kernel_id: Union[None, str] = None
    alive: Union[None, bool] = None
    recorded: Union[None, bool] = None
    managed: Union[None, bool] = None

    @classmethod
    def from_manager(cls, manager: KernelManager) -> "KernelRecord":
        """Create a kernel from a KernelManager."""
        record = cls()
        # Look for the record fields as attributes the kernel manager
        # and make the values match.
        for field in cls.get_identifier_fields():
            setattr(record, field, getattr(manager, field, None))
        record.managed = True
        return record

    @classmethod
    def get_identifier_fields(cls):
        """The identifier keys/labels for a KernelRecord."""
        identifier_fields = []
        for field in fields(cls):
            if field.name.endswith("_id"):
                identifier_fields.append(field.name)
        return identifier_fields

    def get_identifier_values(self):
        """The values of all identifiers."""
        identifier_values = []
        for id in self.get_identifier_fields():
            if id.endswith("_id"):
                identifier_values.append(getattr(self, id))
        return identifier_values

    def get_active_identifiers(self):
        """Return a dictionary of all identifiers that are not None."""
        identifiers = {}
        for id in self.get_identifier_fields():
            val = getattr(self, id)
            if val is not None:
                identifiers[id] = val
        return identifiers

    def __eq__(self, other: object) -> bool:
        """Two kernel records are equivalent if *any* of their
        unique identifiers (keys ending in `_id`) are equal.
        """
        if isinstance(other, KernelRecord):
            current_ids = self.get_active_identifiers()
            incoming_ids = other.get_active_identifiers()

            equivalence_found = False
            conflicts = []

            for field, current_id_value in current_ids.items():
                other_id_value = incoming_ids.get(field, None)
                if other_id_value and other_id_value == current_id_value:
                    equivalence_found = True
                if other_id_value and other_id_value != current_id_value:
                    conflicts.append(field)

            if equivalence_found and conflicts:
                raise KernelRecordConflict(
                    "Two kernel records show some IDs are equivalent, "
                    "while others are not. Here is a list of conflicting "
                    f"IDs: {conflicts}"
                )

            return equivalence_found
        return False

    def update(self, other: "KernelRecord") -> None:
        """Updates in-place a kernel from other (only accepts positive updates"""
        if not isinstance(other, KernelRecord):
            raise TypeError("'other' must be an instance of KernelRecord.")

        if other.kernel_id and self.kernel_id and other.kernel_id != self.kernel_id:
            raise KernelRecordConflict(
                "Could not update the record from 'other' because the two records conflict."
            )

        for field in fields(self):
            if hasattr(other, field.name) and getattr(other, field.name):
                setattr(self, field.name, getattr(other, field.name))


class KernelRecordList:
    """Handy object for storing and managing a list of KernelRecords.
    When adding a record to the list, first checks if the record
    already exists. If it does, the record will be updated with
    the new information.
    """

    def __init__(self, *records):
        self._records: List[KernelRecord] = []
        for record in records:
            self.update(record)

    def __str__(self):
        return str(self._records)

    def __contains__(self, record: Union[KernelRecord, str]) -> bool:
        """Search for records by kernel_id and session_id"""
        if isinstance(record, KernelRecord) and record in self._records:
            return True

        if isinstance(record, str):
            for r in self._records:
                if record in r.get_identifier_values():
                    return True
        return False

    def __len__(self):
        return len(self._records)

    def get(self, record: Union[KernelRecord, str]) -> KernelRecord:
        if isinstance(record, str):
            for r in self._records:
                if record in r.get_identifier_values():
                    return r
        elif isinstance(record, KernelRecord):
            for r in self._records:
                if record == r:
                    return record
        raise ValueError(f"{record} not found in KernelRecordList.")

    def update(self, record: KernelRecord) -> None:
        """Update a record in-place or append it if not in the list."""
        try:
            idx = self._records.index(record)
            self._records[idx].update(record)
        except ValueError:
            self._records.append(record)

    def remove(self, record: KernelRecord) -> None:
        """Remove a record if its found in the list. If it's not found,
        do nothing.
        """
        if record in self._records:
            self._records.remove(record)



import inspect

from traitlets import TraitType


class Awaitable(TraitType):

    info_text = "an awaitable"

    def validate(self, obj, value):
        if not inspect.iscoroutinefunction(value) and not inspect.isawaitable(value):
            raise self.error(obj, value)
        return value



from jupyter_server.base.handlers import APIHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin
from tornado.web import HTTPError


class SynchronizerHandler(ExtensionHandlerMixin, APIHandler):
    async def post(self):
        """Trigger the synchronizer"""
        try:
            await self.extensionapp.sync_managers()
        except Exception as err:
            raise HTTPError(500, log_message=str(err))


handlers = [
    (r"/api/sync", SynchronizerHandler),
]



import asyncio
import pathlib
import uuid

from jupyter_server.extension.application import ExtensionApp
from jupyter_server.utils import run_sync
from traitlets import (
    Bool,
    Float,
    Instance,
    TraitError,
    Type,
    Unicode,
    default,
    validate,
)

from .gateway import fetch_gateway_kernels
from .handlers import handlers
from .kernel_db import KernelTable
from .kernel_records import KernelRecord, KernelRecordList
from .traits import Awaitable


class SynchronizerExtension(ExtensionApp):
    """A Jupyter Server extension for syncing all managers in Jupyter Server."""

    name = "synchronizer"
    handlers = handlers

    autosync = Bool(
        default_value=False,
        help="If True, the extension will periodically synchronize the server automatically.",
    ).tag(config=True)

    syncing_interval = Float(
        default_value=5.0,
        help="Interval (in seconds) for each call to the periodic syncing method.",
    ).tag(config=True)

    database_filepath = Unicode(
        default_value=":memory:",
        help=(
            "The filesystem path to SQLite Database file "
            "(e.g. /path/to/session_database.db). By default, the session "
            "database is stored in-memory (i.e. `:memory:` setting from sqlite3) "
            "and does not persist when the current Jupyter Server shuts down."
        ),
    ).tag(config=True)

    @validate("database_filepath")
    def _validate_database_filepath(self, proposal):
        value = proposal["value"]
        if value == ":memory:":
            return value
        path = pathlib.Path(value)
        if path.exists():
            # Verify that the database path is not a directory.
            if path.is_dir():
                raise TraitError(
                    "`database_filepath` expected a file path, but the given path is a directory."
                )
            # Verify that database path is an SQLite 3 Database by checking its header.
            with open(value, "rb") as f:
                header = f.read(100)

            if not header.startswith(b"SQLite format 3") and not header == b"":
                raise TraitError("The given file is not an SQLite database file.")
        return value

    kernel_record_class = Type(default_value=KernelRecord, klass=KernelRecord).tag(config=True)

    _kernel_records = KernelRecordList()

    kernel_table_class = Type(default_value=KernelTable, klass=KernelTable)
    kernel_table = Instance(klass=KernelTable)

    @default("kernel_table")
    def _default_kernel_remote_table(self):  # pragma: no cover
        return KernelTable()

    fetch_running_kernels = Awaitable(
        help=(
            "The coroutine function used to fetch and record running kernels "
            "that might not be found/managed by Jupyter Server (i.e. they "
            "are managed by a remote Kernel Gateway)."
        )
    ).tag(config=True)

    @default("fetch_running_kernels")
    def default_fetch_running_kernels(self):
        return fetch_gateway_kernels

    multi_kernel_manager = Instance(
        klass="jupyter_server.services.kernels.kernelmanager.MappingKernelManager",
    )

    @default("multi_kernel_manager")
    def _default_multi_kernel_manager(self):
        return self.serverapp.kernel_manager

    contents_manager = Instance(
        klass="jupyter_server.services.contents.manager.ContentsManager",
    )

    @default("contents_manager")
    def _default_contents_manager(self):
        return self.serverapp.contents_manager

    session_manager = Instance(
        klass="jupyter_server.services.sessions.sessionmanager.SessionManager",
    )

    @default("session_manager")
    def _default_session_manager(self):
        return self.serverapp.session_manager

    def fetch_recorded_kernels(self) -> None:
        """Fetch kernels stored in the local Kernel Database."""
        for record in self.kernel_table.list():
            record.recorded = True
            self._kernel_records.update(record)

    def fetch_managed_kernels(self) -> None:
        """Fetch kernel records from any managed kernels (instances of
        KernelManagers) in the MultiKernelManager.
        """
        for km in self.multi_kernel_manager._kernels.values():
            record = self.kernel_record_class.from_manager(km)
            self._kernel_records.update(record)

    async def fetch_kernel_records(self):
        """Fetch all the information that can be found about
        kernels started by this server.
        """
        await self.fetch_running_kernels(self)
        self.fetch_recorded_kernels()
        self.fetch_managed_kernels()
        # Log all kernel records seen at this stage
        self.log.debug(self._kernel_records)

    def record_kernels(self):
        """Record the current kernels to the kernel database."""
        for kernel in self._kernel_records._records:
            conditions = [
                # Kernel isn't already recorded
                not kernel.recorded,
                # Kernel has all identifiers
                all(kernel.get_identifier_values()),
                # Kernel is still running
                kernel.alive,
            ]
            if all(conditions):
                self.kernel_table.save(kernel)
                kernel.recorded = True

    def remove_stale_kernels(self):
        """Remove kernels from the database that are no longer running."""
        for k in self._kernel_records._records:
            if not k.alive:
                self._kernel_records.remove(k)
                if k.recorded:
                    self.kernel_table.delete(kernel_id=k.kernel_id)

    async def hydrate_kernel_managers(self):
        """Create KernelManagers for kernels found for this
        server but are not yet managed.
        """
        for k in self._kernel_records._records:
            if not k.managed and k.alive:
                if not k.kernel_id:
                    kernel_id = str(uuid.uuid4())
                    k.kernel_id = kernel_id
                identifiers = k.get_active_identifiers()
                await self.multi_kernel_manager.start_kernel(**identifiers)
                k.managed = True

    async def delete_stale_sessions(self):
        """Delete sessions that either have no kernel or no content
        found in the server.
        """
        session_list = await self.session_manager.list_sessions()
        mkm = self.multi_kernel_manager
        for session in session_list:
            kid = session["kernel"]["id"]
            known_kids = list(mkm._kernels.keys()) + list(mkm._pending_kernels.keys())
            if kid not in known_kids:
                self.log.debug(
                    f"Kernel {kid} found in the session_manager but "
                    f"not in the kernel_manager. Deleting this session."
                )
                # session = await self.get_session(kernel_id=kid)
                self.session_manager.cursor.execute("DELETE FROM session WHERE kernel_id=?", (kid,))
            # Check the contents manager for documents.
            file_exists = self.contents_manager.exists(path=session["path"])
            if not file_exists:
                session_id = session["id"]
                self.log.debug(
                    f"The document path for {session_id} was not found. Deleting this session."
                )
                await self.session_manager.delete_session(session_id)

    async def shutdown_kernels_without_sessions(self):
        """Shutdown 'unknown' kernels (found in kernelmanager but
        not the session manager).
        """
        for kernel_id in self.multi_kernel_manager.list_kernel_ids():
            try:
                kernel = await self.session_manager.get_session(kernel_id=kernel_id)
            except Exception:
                try:
                    kernel = self.multi_kernel_manager.get_kernel(kernel_id)
                    if (
                        not kernel.ready.done()
                        or kernel_id in self.session_manager._pending_sessions
                    ):
                        continue
                    self.log.debug(
                        f"Kernel {kernel_id} found in the kernel_manager is not "
                        f"found in the session database. Shutting down the kernel."
                    )
                    await self.multi_kernel_manager.shutdown_kernel(kernel_id)
                # Log any failures, but don't raise exceptions.
                except Exception as err2:
                    self.log.info(err2)
                    pass

    async def sync_kernels(self):
        """Synchronize the kernel manager, kernel database, and
        remote kernel service.
        """
        self._kernel_records = KernelRecordList()
        await self.fetch_kernel_records()

        self.remove_stale_kernels()
        await self.hydrate_kernel_managers()
        self.record_kernels()

    async def sync_sessions(self):
        """Synchronize the session database and with the
        multi-kernel_manager by:

        1. Deleting sessions that do not have running
            kernels in the kernel manager
        2. Shutting down kernels in the kernel manager
            that do not have a session associated with them.
        3. Deleting sessions+kernels that do not have content
            found by the contents manager.
        """
        await self.delete_stale_sessions()
        await self.shutdown_kernels_without_sessions()

    async def sync_managers(self):
        """Rehydrate sessions and kernels managers from the remote
        kernel service.
        """
        self.log.debug("Synchronizing kernel records.")
        await self.sync_kernels()
        self.log.debug("Synchronizing kernel sessions.")
        await self.sync_sessions()

    async def _regular_syncing(self, interval=5.0):
        """Start regular syncing on a defined interval."""
        while True:
            self.log.info("Synchonizer is starting another loop.")
            # Try to synchonizer. If failed, log the exception.
            try:
                await self.sync_managers()
            except Exception as err:
                self.log.error(f"Synchonizer failed: {err}")
                if self.log.isEnabledFor(10):
                    self.log.exception(err)
            await asyncio.sleep(interval)

    def start_regular_syncing(self):
        """Run regular syncing in a background task."""
        return asyncio.ensure_future(self._regular_syncing(interval=self.syncing_interval))

    def initialize_settings(self):
        self.initialize_configurables()
        return super().initialize_settings()

    def initialize_configurables(self):
        self.update_config(self.serverapp.config)
        self.kernel_table = self.kernel_table_class(
            database_filepath=self.database_filepath,
            kernel_record_class=self.kernel_record_class,
        )
        # Run the synchronizer once before starting the webapp, to
        # ensure we start with the most accurate state.
        run_sync(self.sync_managers())
        if self.autosync:
            self.start_regular_syncing()



import pathlib
import sqlite3
from typing import Any, List

from traitlets import TraitError, Type, Unicode, validate
from traitlets.config.configurable import Configurable

from .kernel_records import KernelRecord


class KernelTable(Configurable):
    """An SQLite database for recorded kernels in the current server."""

    _table_name = "kerneltable"
    _connection = None
    _cursor = None

    database_filepath = Unicode(
        default_value=":memory:",
        help=(
            "The filesystem path to SQLite Database file "
            "(e.g. /path/to/session_database.db). By default, the session "
            "database is stored in-memory (i.e. `:memory:` setting from sqlite3) "
            "and does not persist when the current Jupyter Server shuts down."
        ),
    ).tag(config=True)

    @validate("database_filepath")
    def _validate_database_filepath(self, proposal):
        value = proposal["value"]
        if value == ":memory:":
            return value
        path = pathlib.Path(value)
        if path.exists():
            # Verify that the database path is not a directory.
            if path.is_dir():
                raise TraitError(
                    "`database_filepath` expected a file path, but the given path is a directory."
                )
            # Verify that database path is an SQLite 3 Database by checking its header.
            with open(value, "rb") as f:
                header = f.read(100)

            if not header.startswith(b"SQLite format 3") and not header == b"":
                raise TraitError("The given file is not an SQLite database file.")
        return value

    kernel_record_class = Type(KernelRecord, klass=KernelRecord)

    @property
    def cursor(self):
        """Start a cursor and create a database called 'session'"""
        if self._cursor is None:
            self._cursor = self.connection.cursor()
            self._cursor.execute(
                f"""CREATE TABLE IF NOT EXISTS {self._table_name}
                ({', '.join(self._table_columns)})"""
            )
        return self._cursor

    @property
    def connection(self):
        """Start a database connection"""
        if self._connection is None:
            self._connection = sqlite3.connect(self.database_filepath, isolation_level=None)
            self._connection.row_factory = sqlite3.Row
        return self._connection

    @property
    def _table_columns(self):
        return set(self.kernel_record_class.get_identifier_fields())

    def query(self, query_string: str, **identifiers: Any) -> None:
        """Build and execute a query."""
        if any(key in identifiers for key in self._table_columns):
            query = query_string.format(
                *list(identifiers.keys()),
                table=self._table_name,
            )
            self.cursor.execute(query, tuple(identifiers.values()))
        else:
            err_message = "A valid identifying field for a Kernel Record was not given."
            identifiers_list = [field for field in self.kernel_record_type if field.endswith("_id")]
            if identifiers_list:
                err_message += f" Examples include: {identifiers_list}"
            raise Exception(err_message)

    def save(self, record: KernelRecord) -> None:
        fields = record.get_active_identifiers()
        columns = ",".join(fields.keys())
        values_tuple = tuple(fields.values())
        if len(values_tuple) > 1:
            values = str(values_tuple)
        else:
            values = f"('{values_tuple[0]}')"
        self.cursor.execute(f"INSERT INTO {self._table_name} ({columns}) VALUES {values}")

    def exists(self, **identifier: Any) -> bool:
        """Check to see if the session of a given name exists"""
        record = self.kernel_record_class(**identifier)
        self.cursor.execute(
            f"SELECT * FROM {self._table_name} WHERE kernel_id='{record.kernel_id}'"
        )
        row = self.cursor.fetchone()
        if row is not None:
            return True
        return False

    def update(self, record: KernelRecord) -> None:
        found = False
        for record_field in record.get_identifier_fields():
            record_id = getattr(record, record_field)
            if record_id and self.exists(**{record_field: record_id}):
                found = True
                break

        if not found:
            raise Exception(
                "No KernelRecord found in the KernelTable. "
                "If this is a new record, use the `.save` method to store "
                "the KernelRecord."
            )

        # Build the query for updating columns.
        values = record.get_active_identifiers()
        updates = []
        for key, value in values.items():
            updates.append(f"{key}='{value}'")
        update_string = ", ".join(updates)
        x = f"UPDATE {self._table_name} SET {update_string} WHERE {record_field}='{record_id}';"
        self.cursor.execute(x)

    def delete(self, **identifier: Any) -> None:
        self.query("DELETE FROM {table} WHERE {0}=?", **identifier)

    def row_to_model(self, row: sqlite3.Row) -> KernelRecord:
        items = {field: row[field] for field in self.kernel_record_class.get_identifier_fields()}
        return self.kernel_record_class(**items)  # type:ignore[no-any-return]

    def list(self) -> List[KernelRecord]:
        self.cursor.execute(f"SELECT * FROM {self._table_name}")
        rows = self.cursor.fetchall()
        return [self.row_to_model(row) for row in rows]

    def get(self, **identifier: Any) -> KernelRecord:
        self.query("SELECT * FROM {table} WHERE {0}=?", **identifier)
        row = self.cursor.fetchone()
        if not row:
            raise Exception("No match was found in database.")
        return self.row_to_model(row)



from dataclasses import dataclass
from typing import Union

import pytest

from jupyter_server_synchronizer.kernel_records import (
    KernelRecord,
    KernelRecordConflict,
    KernelRecordList,
)


def test_kernel_record():
    assert KernelRecord.get_identifier_fields() == ["kernel_id"]

    record1 = KernelRecord(kernel_id="kernel1")

    assert record1.get_identifier_values() == ["kernel1"]


def test_kernel_record_equals():
    record0 = KernelRecord(kernel_id="kernel1")
    record1 = KernelRecord(kernel_id="kernel2")
    record2 = KernelRecord(kernel_id="kernel1", alive=True)
    record3 = KernelRecord(kernel_id="kernel2", alive=True)

    assert record0 != record1
    assert record0 == record2
    assert record0 != record3


def test_kernel_record_update():
    record = KernelRecord(kernel_id="kernel1")
    record_update = KernelRecord(kernel_id="kernel1", alive=True)
    record_update2 = KernelRecord(kernel_id="kernel1", recorded=True)

    record.update(record_update)
    assert record.alive

    record.update(record_update2)
    assert record.recorded


def test_kernel_record_list():
    records = KernelRecordList()
    r = KernelRecord(kernel_id="kernel1")
    records.update(r)
    assert r in records
    assert "kernel1" in records
    assert len(records) == 1

    # Test .get()
    r_ = records.get(r)
    assert r == r_
    assert r.kernel_id is not None
    r_ = records.get(r.kernel_id)
    assert r == r_

    with pytest.raises(ValueError):
        records.get("badkernel")

    r_update = KernelRecord(kernel_id="kernel1", alive=True)
    records.update(r_update)
    assert len(records) == 1
    assert "kernel1" in records
    assert records.get("kernel1").alive

    r2 = KernelRecord(kernel_id="kernel2")
    records.update(r2)
    assert r2 in records
    assert len(records) == 2

    records.remove(r2)
    assert r2 not in records
    assert len(records) == 1


@dataclass(eq=False)
class CustomKernelRecord(KernelRecord):
    remote_id: Union[None, str] = None


def test_custom_kernel_record():
    assert CustomKernelRecord.get_identifier_fields() == ["kernel_id", "remote_id"]


def test_custom_kernel_record_equals():
    record0 = CustomKernelRecord(kernel_id="kernel1")
    record1 = CustomKernelRecord(remote_id="remote1")
    record2 = CustomKernelRecord(remote_id="remote1", kernel_id="kernel1")
    record3 = CustomKernelRecord(remote_id="remote2", kernel_id="kernel1")
    record4 = CustomKernelRecord(remote_id="remote1", kernel_id="kernel2")
    record5 = CustomKernelRecord(remote_id="remote1", kernel_id="kernel1")

    assert record0 == record2
    assert record1 == record2
    assert record3 != record4
    assert record1 != record3
    assert record3 != record4
    assert record2 == record5

    with pytest.raises(KernelRecordConflict):
        assert record2 == record3

    with pytest.raises(KernelRecordConflict):
        assert record2 == record4


def test_custom_kernel_record_update():
    record1 = CustomKernelRecord(remote_id="remote1")
    record2 = CustomKernelRecord(remote_id="remote1", kernel_id="kernel1")
    record1.update(record2)
    assert record1.kernel_id == "kernel1"

    record1 = CustomKernelRecord(remote_id="remote1")
    record2 = CustomKernelRecord(kernel_id="kernel1")
    record1.update(record2)
    assert record1.kernel_id == "kernel1"

    record1 = CustomKernelRecord(kernel_id="kernel1")
    record2 = CustomKernelRecord(remote_id="remote1")
    record1.update(record2)
    assert record1.remote_id == "remote1"

    record1 = CustomKernelRecord(kernel_id="kernel1")
    record2 = CustomKernelRecord(remote_id="remote1", kernel_id="kernel1")
    record1.update(record2)
    assert record1.remote_id == "remote1"

    record1 = CustomKernelRecord(kernel_id="kernel1")
    record2 = CustomKernelRecord(remote_id="remote1", kernel_id="kernel2")
    with pytest.raises(KernelRecordConflict):
        record1.update(record2)

    record1 = CustomKernelRecord(kernel_id="kernel1", remote_id="remote1")
    record2 = CustomKernelRecord(kernel_id="kernel2")
    with pytest.raises(KernelRecordConflict):
        record1.update(record2)

    record1 = CustomKernelRecord(kernel_id="kernel1", remote_id="remote1")
    record2 = CustomKernelRecord(kernel_id="kernel2", remote_id="remote1")
    with pytest.raises(KernelRecordConflict):
        record1.update(record2)

    record1 = CustomKernelRecord(remote_id="remote1", kernel_id="kernel1")
    record2 = CustomKernelRecord(remote_id="remote2", kernel_id="kernel1")
    record1.update(record2)
    assert record1.remote_id == "remote2"


def test_kernel_record_list_of_custom_records():
    records = KernelRecordList()
    r = CustomKernelRecord(kernel_id="kernel1")
    records.update(r)
    assert r in records
    assert "kernel1" in records
    assert len(records) == 1

    # Test .get()
    r_ = records.get(r)
    assert r == r_
    assert r.kernel_id is not None
    r_ = records.get(r.kernel_id)
    assert r == r_

    with pytest.raises(ValueError):
        records.get("badkernel")

    r_update = CustomKernelRecord(kernel_id="kernel1", remote_id="remote1")
    records.update(r_update)
    assert len(records) == 1
    assert "remote1" in records

    r2 = CustomKernelRecord(kernel_id="kernel2")
    records.update(r2)
    assert r2 in records
    assert len(records) == 2

    records.remove(r2)
    assert r2 not in records
    assert len(records) == 1
